/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file audio.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 25.02.2006
 */


/**
 * Includes
 */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>
#include <stdbool.h>

#include "common.h"		/** common project settings */
#include "audio.h"		/** common for all audio */
#include "mp3.h"		/** mp3 decoder */
#include "parsetime.h"	/** converts time-formats */
#include "report.h"		/** reporter class */

#ifdef WITH_AAC
#include "aac.h"		/** aac decoder */
#endif	// WITH_AAC


/*
 * Global vars
 */
static
audio_data_t *data = NULL;	/** the main data struct. this contains all information relating to the audio file and playback */

static
char *prev_filename = NULL;	/** the name of the previous file we've played. used to make sure we don't open files more than we have to */

static
bool is_playing;			/** tells us if we're currently playing a file.
								(this used to be in data struct, but is only
								referenced in this file, so we move it out
								of the struct) */


/**
 * Function declarations
 */
static
bool open_file (const char *fname, audio_data_t *data);

static
audio_data_t *audio_data_init ();

static
audio_file_t get_file_type (const char *file);

static
bool is_new_file (audio_data_t *data);

static
void (*playing_done)(void *arg) = NULL;

/**
 * Function definitions
 */
/**
 * Finds out the type of file and starts the aproporiate decoder
 * 
 * data should be mutex locked when this function is called.
 */
void audio_decoder ()
{
	if (DEBUG_AUDIO) report ("playing file. . ", REP_DEBUG);

	switch (get_file_type(data->filename))	/** finds out the filetype */
	{
#ifdef WITH_AAC	/** do we build with or without AAC support? */
		case (AUDIO_FILE_AAC):
			if (DEBUG_AUDIO) report ("audio.c: audio-file is aac", REP_DEBUG);

//			pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */

			if ((is_new_file (data)) &&
				(data->prev_file_type != AUDIO_FILE_AAC))	/** last file_type was NOT aac */
			{
				if (DEBUG_AUDIO) report ("the last filetype was NOT aac", REP_DEBUG);
				mp3_close (data);	/** close decoder and start a new one */
				aac_init (data);
			}
			(void *) aac_play (data);

//			pthread_mutex_unlock (&data->mutex);	/** release mutex - we'rse done with "data" */

			break;
#endif	// WITH_AAC
		case (AUDIO_FILE_MP3):
			if (DEBUG_AUDIO) report ("audio.c: audio-file is mp3", REP_DEBUG);

//			pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */

			if ((is_new_file (data)) &&
				(data->prev_file_type != AUDIO_FILE_MP3)) /** last file_type was NOT mp3 */
			{
				if (DEBUG_AUDIO) report ("the last filetype was NOT mp3", REP_DEBUG);
#ifdef WITH_AAC
				aac_close (data);	/** close decoder and start a new one */
#endif	// WITH_AAC
				mp3_init (data);
			}
			(void *) mp3_play (data);

//			pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */

			break;
		case (AUDIO_FILE_WAV):
			report ("wav parsing isn't implemented yet", REP_MESSAGE);
			break;
		case (AUDIO_FILE_UNKNOWN):
		default:		/** This should never happen! */
			assert (0);	/** Report where we are and exit the program */
			break;
	}

	if (DEBUG_AUDIO) report ("start_decoder thread exiting", REP_DEBUG);

	is_playing = false;			/** we're not playing anymore */
	data->state = AUDIO_STATE_STOP;		/** settings state to stopped */
	pthread_cond_signal (&data->finished);	/** signal that we are finished playing a segment */

	playing_done (NULL);	/** calling callback function. this will let the
								frontend know that we are done playing */

	pthread_exit (0);
}


/**
 * Plays a file from start to stop.
 * returns the duration in milliseconds
 */
unsigned
long
int audio_play (const char *fname, char *start, char *stop)
{
	sched_yield();	/** We call sched_yield here to give the other threads
						audio_block a chance to catch up with the change in
						data->state. */
	
	pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */

	if (data->state != AUDIO_STATE_STOP) {	/** make sure we are stopped */
		report ("audio_play called when state is NOT stopped", REP_ERROR);
	}

	data->progress.seconds	= 0;	/** reset the progress */
	data->progress.fraction	= 0;

	data->start	=	parse_time_mad_time_t (start);	/** and set start- and stop-times. */
	data->stop	=	parse_time_mad_time_t (stop);
	int duration = parse_time_get_dur(start, stop);	/** get duration in ms */
	
	if (open_file (fname, data) == false) {	/** constructing and reporting an error-message */
		char message[STRLEN];
		snprintf (message, sizeof (message), "Could not open file. (%s)", fname);
		report (message, REP_WARNING);
		return false;
	}

	data->state = AUDIO_STATE_PLAY;		/** set state to playing */

	pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
	
	sched_yield();	/** we must let go of the CPU!
						if the caller calls audio_block before the
						AUDIO_STATE_PLAY are acted upon, we get weird errors.
						we must give the other threads a chance to catch up. */

	return duration;	/** we return the duration of this clip */
}


/**
 * audio_block () will block for as long as the engine outputs audio.
 * if this is called right after audio_play (), it will block the caller
 * until the audio-segment is finished.
 */
void audio_block ()
{
	sched_yield();	/** TODO: Evaluate whether this prevents audio_block from
						ending prematurelly. possibly because is_playing
						isn't set correctly */
	if (is_playing)
	{
		if (DEBUG_AUDIO) report ("entering audio_block", REP_DEBUG);
		pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */
		pthread_cond_wait(&data->finished, &data->mutex);
		pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
		if (DEBUG_AUDIO) report ("exiting audio_block", REP_DEBUG);
	}
	else
		if (DEBUG_AUDIO) report ("audio_block called when we're not playing", REP_DEBUG);
//		pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
}


/**
 * A thread function. This is run as a thread from daisyplayer.c
 * and keeps the audio-engine alive.
 */
void audio_thread ()
{
	if (DEBUG_AUDIO) report ("in audio_thread", REP_DEBUG);

	pthread_t audio_decoder_thread;

	pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */
	is_playing = false;
	pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */

	while (1)
	{
		pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */

		if (data->state == AUDIO_STATE_PLAY)
		{
			if (!is_playing)
			{
				if (DEBUG_AUDIO) report ("audio_thread: starting start_decoder thread", REP_DEBUG);
				pthread_create(&audio_decoder_thread, NULL, (void *) audio_decoder, NULL);
				is_playing = true;
			}
		}
		if ((data->state == AUDIO_STATE_STOP) && (is_playing))
		{
			if (DEBUG_AUDIO) report ("audio_thread: stopping start_decoder thread", REP_DEBUG);
			pthread_cancel(audio_decoder_thread);	/** stopping thread */
			is_playing = false;						/** we're not playing anymore */
		}

		pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
		sched_yield();	/** give up the cpu */
	}
}


/**
 * Sets the state to stop
 */
void audio_stop ()
{
//	if (is_playing)
//	{
//		printf ("\nstate: %d\tis_playing: %d", data->state, is_playing);
		if (DEBUG_AUDIO) report ("stopping", REP_DEBUG);
		pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */
		data->state = AUDIO_STATE_STOP;
		is_playing = false;
		pthread_cond_signal(&data->finished);	/** signal to audio_block that we are finished */
		pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
//		printf ("\nstate: %d\tis_playing: %d", data->state, is_playing);
//	}
}


/**
 * Sets the state to pause
 */
void audio_pause ()
{
	if (is_playing)
	{
		if (data->state == AUDIO_STATE_PAUSE)
		{
			if (DEBUG_AUDIO) report ("unpausing", REP_DEBUG);
			pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */
			data->state = AUDIO_STATE_PLAY;
			pthread_cond_signal (&data->unpause);
			pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
		}
		else
		{
			if (DEBUG_AUDIO) report ("pausing", REP_DEBUG);
			pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */
			data->state = AUDIO_STATE_PAUSE;
			pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
		}
	}
}

/**
 * callback should be called by the decoder loop
 */
int callback ()
{
	int ret = 0;
	pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */
	if (data->state == AUDIO_STATE_PAUSE)
		pthread_cond_wait (&data->unpause, &data->mutex);
	if (data->state == AUDIO_STATE_STOP)
		ret = -1;	/** -1 tells the decoder to stop */
	pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
	return ret;
}


/**
 * true if the current file is difrent from
 * the last time this function was called
 * 
 * data should be under a mutex lock when this method is called.
 */
static
bool is_new_file (audio_data_t *data)
{
	if (prev_filename != NULL)	/** prev_filename is already set to a value */
	{
//		pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */
		if (!strcmp (prev_filename, data->filename))
		{
//			pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
			return false;	/** the filenames are the same */
		}
		else
		{
//			pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
			return true;	/** the filenames differ */
		}
	}
	else
	{
//		pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */
		prev_filename = data->filename;
//		pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
		return true;
	}
}


/**
 * Creates a new audio_data_t struct
 */
static
audio_data_t *audio_data_init ()
{
	audio_data_t *data;
	data = (audio_data_t *) malloc (sizeof (audio_data_t));
	if (data == NULL) return NULL;

	/** And setting some values to safe defaults */
	data->fdm = NULL;
	data->filename = NULL;	

	data->file_type = AUDIO_FILE_UNKNOWN;
	data->prev_file_type = AUDIO_FILE_UNKNOWN;

	data->state = AUDIO_STATE_STOP;		/** default state is stopped */
	
	/** Setting up some thread data */
	if (pthread_cond_init(&data->unpause, NULL) != 0)
		report ("Could not initialize data->unpause", REP_ERROR);
	if (pthread_cond_init(&data->finished, NULL) != 0)
		report ("Could not initialize data->finished", REP_ERROR);
	if (pthread_mutex_init(&data->mutex, NULL) != 0)
		report ("Could not initialize data->mutex", REP_ERROR);

	if (DEBUG_AUDIO) report ("data struct initialized", REP_DEBUG);
	return (data);
}


/**
 * This function will load a file if it's not already loaded.
 * The freeing and generating of mmaps are handled here.
 * Returns -1 on error, 1 on success.
 * data should be mutex locked when this function is called
 */
static
bool open_file (const char *fname, audio_data_t *data)
{
	if (DEBUG_AUDIO) {
		char rep[STRLEN];
		snprintf (rep, sizeof (rep), "audio.c asked to open file: %s\t(data->filename is: %s)", fname, data->filename);
//		report (rep, REP_DEBUG);
	}

	/** did we already load the file? */
	/** We have to check if data->filename is NULL /first!/ If one of
	 * strcmp's arguments are NULL, we've made a booboo */
	if ((data->filename != NULL) && !strcmp (data->filename, fname))
	{
		if (DEBUG_AUDIO) report ("File already loaded", REP_DEBUG);
	}

	/** if not, we'll need to free the old mmap, open the new
	 * file and create a new mmap */
	else	/** file is NOT already loaded */
	{
		/* Free mmap, clean data */
		if (data->fdm != NULL) {

			/** freeing the mmap before creating a new one */
			if (munmap (data->fdm, data->stat.st_size) == -1)
				report ("Unable to free mmap", REP_CRITICAL);

			if (DEBUG_AUDIO) {
				char rep[STRLEN];
				snprintf (rep, sizeof (rep), "Freed mmap. Pointer: %p, Size: %d", data->fdm, (int) data->stat.st_size);
				report (rep, REP_DEBUG);
			}

			data->fdm = NULL;		/** we have freed the data - don't want an old pointer to point there! */
			data->filename = NULL;	/** this file is no longer in our memory */
			close (data->fd);		/** closing the file before we open a new one */
		}
		
		/** udating the data-struct with the new filename */
		if (data->filename != NULL)
			free(data->filename);
		data->filename = (char *) malloc (strlen(fname)+1);
		if (DEBUG_AUDIO) report ("allocating room for data->filename", REP_DEBUG);
		strcpy(data->filename, fname);
		if (DEBUG_AUDIO) report ("copying fname to data->filename", REP_DEBUG);
		char msg[STRLEN];
		snprintf (msg, sizeof (msg), "data->filename is: %s\tfname is: %s", data->filename, fname);
		if (DEBUG_AUDIO) report (msg, REP_DEBUG);


		/** opening new file */
		if ((data->fd = open (fname, O_RDONLY)) == -1)
		{
			report ("Error opening file", REP_WARNING);
			return false;
		}

		/** setting filetype */
		data->prev_file_type = data->file_type;
		data->file_type = get_file_type(fname);

		/** fstating file. this gives us alot of information about
		 * the file (we'll use the st_size (filesize) here) */
		if (stat(fname, &data->stat) == -1 || data->stat.st_size == 0)
		{
			report ("Error fstating file or filesize is 0", REP_WARNING);
			return false;
		}

		/** mmaps the file */
		data->fdm = mmap (0, data->stat.st_size, PROT_READ, MAP_SHARED, data->fd, 0);
		if (data->fdm == MAP_FAILED)
		{
			report ("Error mmapping file", REP_WARNING);
			return false;
		}
		
		if (DEBUG_AUDIO)
		{
			char rep[STRLEN];		
			snprintf (rep, sizeof (rep), "mmapped new file. Pointer: %p, size: %d", data->fdm, (int) data->stat.st_size);		
			report (rep, REP_DEBUG);
		}
	}

	return true;
}



/*
 * Finds out what filetype a file is (mp3, aac, wav).
 * TODO: Make this come up with an intelligent guess at least
 * wether the file is mp3 or mp4/aac.
 */
static
audio_file_t get_file_type (const char *fname)
{
//	if (DEBUG_AUDIO) report("Querying file_type. Returning mp3 for now.", REP_DEBUG);
	return AUDIO_FILE_MP3;
}


/**
 * This is called when the program wants us to stop and free our resources.
 * This should make sure resources are freed.
 */
void audio_terminate ()
{
	pthread_mutex_lock (&data->mutex);	/** get mutex - we work on "data" */

	data->state = AUDIO_STATE_STOP;
	sched_yield();

	if (data != NULL)	/** If data is NULL, then it's never been */
	{					/** used and we don't have to do anything */
		switch (data->file_type)
		{
			case (AUDIO_FILE_MP3):
				mp3_close(data);
				break;
			case (AUDIO_FILE_AAC):
#ifdef WITH_AAC
				aac_close(data);
#endif	// WITH_AAC
				break;
			case (AUDIO_FILE_WAV):
//				wav_close(data);
				break;
			default:
				break;
		}

		pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
		free (data);	//Free data struct
	}
	else
	{
		if (DEBUG_AUDIO) report ("audio_terminate called before audio are initialized", REP_DEBUG);
	}
}


/**
 * This is called when the program wants us to start and prepare for audio playback
 * This should allocate the resources we need.
 */
void audio_initiate (void (*cb)(const void *))
{
	playing_done = cb;		/** callback function. playback_done () should
								be called whenever we're done playing
								a segment */
	data = audio_data_init();	/** initiate data structure */
}
