/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file callback.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 * 
 * Datastructure and methods for controlling it.
 */

 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <libxml2/libxml/xmlreader.h>

#include "control.h"
#include "nccparser.h"
#include "ncxparser.h"
#include "nccsmilparser.h"
#include "ncxsmilparser.h"
#include "report.h"
#include "common.h"
#include "audio.h"

#ifdef LIBXML_READER_ENABLED

// static function declarations
static void initNode (struct Node* node);
static void initSmilNode (struct SmilNode* smilNode);
static void flushSmilNodes ();
static void flushNodes ();


/**
 * Initializes the data structure for the parsing (linked lists)
 */
void parseInit ()
{
	 /*
     * this initialize the library and check potential ABI mismatches
     * between the version it was compiled for and the actual shared
     * library used.
     */
	LIBXML_TEST_VERSION
	
	/* initalizing the datastucture for storing book meta data */
	__bookInfo = (struct BookInfo *) malloc (sizeof (struct BookInfo));
	__bookInfo->titleText = NULL;
	__bookInfo->titleAudiofilename = NULL;
	__bookInfo->titleAudioStartPos = NULL;
	__bookInfo->titleAudioStopPos = NULL;
	__bookInfo->titleImage = NULL;

	/* initializing the head and the tail of the linked lists */
	__smilHead = (struct SmilNode *) malloc (sizeof (struct SmilNode));
	initSmilNode (__smilHead);
	__smilTail = (struct SmilNode *) malloc (sizeof (struct SmilNode));
	initSmilNode (__smilTail);
	__smilHead->next = __smilTail;
	__smilTail->prev = __smilHead;
	__smilPos = __smilHead;
	
	__nodeHead = (struct Node *) malloc (sizeof (struct Node));
	initNode (__nodeHead);
	__nodeTail = (struct Node *) malloc (sizeof (struct Node));
	initNode (__nodeTail);
	__nodeHead->next = __nodeTail;
	__nodeTail->prev = __nodeHead;
	__nodePos = __nodeHead;
}


/**
 * Method to free all memory allocated by the parser. Terminates all parsing
 * functionality.
 */
void parseTerminate ()
{
	flushNodes();
	flushSmilNodes();
	
	/*
	* Cleanup function for the XML library.
	*/
	xmlCleanupParser ();
    
	/*
	 * this is to debug memory for regression tests
	*/
	xmlMemoryDump ();
	
	free (__nodeHead);
	free (__nodeTail);	
}


/**
 * Removes whitespaces, newlines and tabs before and after a sentence.
 *
 * @param string contains the string to check.
 *
 * Return the new char* if the sentence contains any characters,
 * NULL otherwise
 */
char* removewhitespaces(char* string)
{	
	int start = 0; // Counter variables	
	int end = 0;
	int i = 0;
	int n = 0;
	char* newstring = NULL; // The new string
	
	/* Checks the characters before the string */
	for (start = 0; start < strlen(string); start++)
	{	
		if (string[start] != ' ' && string[start] != '\t' && string[start] != '\n')
			break;
	}
	
	/* If the sentences only contains whitespaces,tabs or newlines, return NULL */
	if (start == strlen(string)) return NULL;
	
	/* Checks the characters from the end and backwards */
	for (end = strlen(string)-1; end >= 0; end--)
	{	
		if (string[end] != ' ' && string[end] != '\t' && string[end] != '\n')
			break;
	}
	
	/* Creates the new string with the non-characters stripped */	
	newstring = malloc(end-start+2);
	for (i = start; i <= end; i++)
	{
		newstring[n++] = string[i];
	}
	newstring[n] = '\0';	
	return newstring;
}


/**
 * Converts min:sec.frac to sec.frac
 * 
 * @return the string contianing the offset
 */
char* getTime(char* time)
{
	char* tmp, *min, *sec, *frac = NULL;
	char* value = (char *) malloc(6);	
	int minutes = 0, seconds = 0;
	
	// copy time to a tmp variable
	tmp = (char *) malloc (strlen (time)+1);
	strcpy (tmp, time);
	// tokenize in order to get minutes
	min = strtok (tmp, ":");
	minutes = atoi (min);
	// tokenize in order to get the chars after min:
	tmp = strtok (NULL, ":");
	// tokenize in order to get seconds
	sec = strtok (tmp, ".");
	// convert char to int in order to add minutes to seconds
	seconds = atoi(sec);
	seconds = (minutes * 60) + seconds;
	// tokenize in order to get fraction
	frac = strtok (NULL, ".");	
	// convert the int seconds to char
	snprintf(value, sizeof (value), "%d", seconds);
	// add .fractions
	strcat (value, ".");
	strcat (value, frac);
//	free (tmp);
	return value;
}


/**
 * Adds a new SmilNode at the end of the linked list
 */
struct SmilNode* addNewSmilNode ()
{
	// allocating memory for the new smilnode
	struct SmilNode *smilNode = (struct SmilNode *) malloc (sizeof(struct SmilNode));
	
	// initializes the smilNode
	initSmilNode (smilNode);
	
	// adding the empty smilnode til the end of the linked list
	struct SmilNode *tmp = __smilTail->prev;
	tmp->next = smilNode;
	smilNode->prev = tmp;
	smilNode->next = __smilTail;
	__smilTail->prev = smilNode;
	
	return smilNode;
}

/**
 * Sets the smilNodes pointers to NULL and all values to 0
 */
static void initSmilNode (struct SmilNode* smilNode)
{
	smilNode->id = NULL;
	smilNode->anchor = NULL;
	smilNode->fragmentIdentifier = NULL;
	smilNode->textpassage = NULL;
	smilNode->audiofilename = NULL;
	smilNode->audioStartPos = NULL;
	smilNode->audioStopPos = NULL;
	smilNode->image = NULL;
	smilNode->header = NULL;
	smilNode->next = NULL;
	smilNode->prev = NULL;
	smilNode->optional = 0;
}

 
 /**
  * Adds a new Node at the end of the linked list
  */
struct Node* addNewNode ()
{
	// allocating memory for the new smilnode
	struct Node *node = (struct Node *) malloc (sizeof(struct Node));
	
	// initializes the node
	initNode(node);
	
	// adding the empty smilnode til the end of the linked list
	struct Node *tmp = __nodeTail->prev;
	tmp->next = node;
	node->prev = tmp;
	node->next = __nodeTail;
	__nodeTail->prev = node;
	
	return node;
}

/**
 * Sets the nodes pointers to NULL and all values to 0
 */
static void initNode (struct Node* node)
{
	node->audiofilename = NULL;
	node->audioStartPos = NULL;
	node->audioStopPos = NULL;
	node->fragmentidentifier = NULL;
	node->image = NULL;
	node->textPassage = NULL;
	node->textfilename = NULL;
	node->next = NULL;
	node->prev = NULL;
}


/**
 * Removes all smilnodes in the linked list and deallocates their memory.
 */
static void flushSmilNodes ()
{
	struct SmilNode *tmp = __smilHead->next;
	while (tmp != __smilTail)
	{
		if (tmp->id != NULL) 
			free (tmp->id);
		if (tmp->anchor != NULL)
			free (tmp->anchor);
		if (tmp->fragmentIdentifier != NULL)
			free (tmp->fragmentIdentifier);
		if (tmp->header != NULL)
			free (tmp->header);
		tmp = tmp->next;
		free (tmp->prev);
	}
	__smilHead->next = __smilTail;
	__smilTail->prev = __smilHead;
	__smilPos = __smilHead;
}


/**
 * Removes all nodes in the linked list and deallocates their memory.
 */
static void flushNodes ()
{
	struct Node *tmp = __nodeHead->next;
	while (tmp != __nodeTail)
	{
		if (tmp->audiofilename != NULL)
			free (tmp->audiofilename);
		if (tmp->audioStartPos != NULL)
			free (tmp->audioStartPos);
		if (tmp->audioStopPos != NULL)
			free (tmp->audioStopPos);
		if (tmp->textPassage != NULL)
			free (tmp->textPassage);
		tmp = tmp->next;
		free (tmp->prev);	
	}
	__nodeHead->next = __nodeTail;
	__nodeTail->prev = __nodeHead;
	__nodePos = __nodeHead;
}


/**
 * Function for seeking in the data structure. It can seek paragraphs (nodes) and
 * chapters (smilnode).
 * 
 * @param type -2 = prev para, -1 = prev chap, 1 = next chap, 2 = next para
 * @param book 0 = ncc (2.02), 1 = ncx (Z39.86-2005)
 * @returns 0 if end of book, 1 if success and -1 on error
 */
int seek (int type, int book)
{
	int ret = 1;
	
	if (__smilHead->next == __smilTail) return 0;
	
	switch (type)
	{
		case -2:
			if (__nodePos == __nodeHead || __nodePos == __nodeHead->next)
			{
				if (__smilPos == __smilHead)
				{
					__smilPos = __smilHead->next;
					flushNodes();
					if(book == 0)
						ret = parseNCCSmil(__smilPos);	
					else
						ret = parseNCXSmil(__smilPos);
					if (ret != 1) return ret;
					__nodePos = __nodeHead->next;
					return ret;	
				}
				else if (__smilPos == __smilHead->next)	
				{
					__nodePos = __nodeHead->next;
					return ret;	
				}
				else
				{
					__smilPos = __smilPos->prev;
					flushNodes();
					if(book == 0)
						ret = parseNCCSmil(__smilPos);	
					else
						ret = parseNCXSmil(__smilPos);
					if (ret != 1) return ret;
					__nodePos = __nodeTail->prev;
					return ret;	
				}
			}
			else
			{
				__nodePos = __nodePos->prev;
				return ret;
			}
		
		case -1:
			if (__smilPos == __smilHead)
			{
				__smilPos = __smilHead->next;
				flushNodes();
				if(book == 0)
					ret = parseNCCSmil(__smilPos);	
				else
					ret = parseNCXSmil(__smilPos);
				if (ret != 1) return ret;
				__nodePos = __nodeHead->next;
				return ret;
			}
			else if (__smilPos == __smilHead->next)
			{
				__nodePos = __nodeHead->next;
				return ret;	
			}
			else
			{
				__smilPos = __smilPos->prev;
				flushNodes();
				if(book == 0)
					ret = parseNCCSmil(__smilPos);	
				else
					ret = parseNCXSmil(__smilPos);
				if (ret != 1) return ret;
				__nodePos = __nodeHead->next;
				return ret;	
			}
		
		case 1:
			if (__smilPos == __smilTail)
			{
				return 0;
			}
			else if (__smilPos == __smilTail->prev)
			{
				__smilPos = __smilTail;
				return 0;
			}
			else {
				__smilPos = __smilPos->next;
				flushNodes();
				if(book == 0)
					ret = parseNCCSmil(__smilPos);	
				else
					ret = parseNCXSmil(__smilPos);
				if (ret != 1) return ret;
				__nodePos = __nodeHead->next;
				return ret;
			}
		
		case 2:
			if (__nodePos == __nodeTail || __nodePos == __nodeTail->prev)
			{
				if (__smilPos == __smilTail)
				{
					__nodePos = __nodeTail;
					return 0;
				}
				else if (__smilPos == __smilTail->prev)
				{
					__smilPos = __smilTail;
					__nodePos = __nodeTail;
					return 0;	
				}
				else
				{
					__smilPos = __smilPos->next;
					flushNodes();
					if(book == 0)
						ret = parseNCCSmil(__smilPos);	
					else
						ret = parseNCXSmil(__smilPos);
					if (ret != 1) return ret;
					__nodePos = __nodeHead->next;
					return ret;
				}
			}
			else
			{
				__nodePos = __nodePos->next;
				return ret;	
			}
		
		default:
			return -1;	
		
	}
}


/**
 * Start parsing a NCC or a NCX file.
 * 
 * @param filenam containing the path/filename to the ncc/ncx file
 * 
 * @returns 1 in case of success and -1 in case of error
 */
int parse(char* filename)
{
	char buffer[STRLEN];
	int i, j, n, ret;
	j = n = ret = 0;
	i = (int) strlen(filename);
	
	// checks wheter the filename ends with .ncx or .NCX
	if((filename[i-1] == 'x' && filename[i-2] == 'c' && filename[i-3] == 'n' && filename[i-4] == '.') 
	|| (filename[i-1] == 'X' && filename[i-2] == 'C' && filename[i-3] == 'N' && filename[i-4] == '.'))
	{	
		// parse NCX, must extract the path
		__ncx = 1;
		// find the path
		for(i = (int) strlen(filename)-1; i > 0; i--)
		{
			if(filename[i] == '/')
				break;
		}
		// stores the path
		for (j = 0; j <= i; j++)
		{
			buffer[n++] = filename[j];
		}
		buffer[n] = '\0';
		__path = (char*) malloc (sizeof (buffer)+1);
		assert (__path != NULL);	/** Make sure path isn't NULL */
		strcpy (__path, buffer);
		ret = parseNCX (filename);	
	}
	// 2.02 Daisy book must be ../ncc.html or ../NCC.HTML
	else if((filename[i-1] == 'l' && filename[i-2] == 'm' && filename[i-3] == 't' && filename[i-4] == 'h'
	&&	filename[i-5] == '.' && filename[i-6] == 'c' && filename[i-7] == 'c' && filename[i-8] == 'n' && filename[i-9] == '/') 
	|| (filename[i-1] == 'L' && filename[i-2] == 'M' && filename[i-3] == 'T' && filename[i-4] == 'H'
	&&	filename[i-5] == '.' && filename[i-6] == 'C' && filename[i-7] == 'C' && filename[i-8] == 'N' && filename[i-9] == '/'))
	{
		// we got a 2.02 Daisy book -> parseNCC
		__ncx = 0;
		
		// stores the path
		for (j = 0; j <= (int) strlen(filename)-9; j++)
		{
			buffer[n++] = filename[j];
		}
		buffer[n] = '\0';
		__path = (char*) malloc (sizeof (buffer)+1);
		assert (__path != NULL);	/** make sure path isn't NULL */
		strcpy (__path, buffer);
		ret = parseNCC (filename);
	}
	else
	{
		printf ("\n\nSorry, you must enter a valid path for the Daisybook (../ncc.html | ../*.ncx)\n\n");
		return -1;
	}
	return ret;
}


/**
 * Starts the audio. This should run as a thread from daisyplayer.c
 */
void start_audio (void (*cb)(const void *))
{
	printf ("in start_audio\n");

	audio_initiate (cb);	/** initiate audio before we use it! */
	
	int ret = pthread_create (&__thread_audio, NULL, (void *) audio_thread, NULL);
	if (ret) {
		report ("Error creating audio thread", REP_CRITICAL);
	}
	else {
		if (DEBUG_DAISYPLAYER) report ("cotrol.c: audio_thread started", REP_DEBUG);
	}
}


void stop_audio ()
{
	audio_terminate();		/** tell audiomodule to terminate and clean up	*/
	pthread_cancel (__thread_audio);
}

#endif
