/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file parsetime.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 15.02.2006
 * 
 * This library parses valid DAISY time formats and constructs mad_timer_t
 * You should probably read section 7.7 "Media Clipping and Clock Values"
 * in ANSI/NISO Z39.86.2005 to get an understanding of the formats we're
 * dealing with.
 */


/**
 * includes
 */
#include <mad.h>	/** we return mad_timer_t structs */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include "common.h"		/** common project settings */
#include "parsetime.h"
#include "report.h"		/** report messages */
#include "common.h"


static enum pt_type type (char* expr);


/*
 * Test-function
 */
//int main (int argc, char *argv[])
//{
//	char expr[] = "1:1:1.100";
////	char expr[] = "43:12.045";
////	char expr[] = "34.6s";
//	mad_timer_t mad_time;
//	mad_time = parse_time_mad_time_t (expr);
//	printf ("mad_timer_t: %li:%lo\n", mad_time.seconds, mad_time.fraction);
//	return 1;
//}


/**
 * simply checks is a char is a letter or not
 */
static
bool isletter (char c)
{
	if ((c > 65) && (c < 90))	/** A-Z */
		return true;
	if ((c > 97) && (c < 122))	/** a-z */	
		return true;
	return false;
}


/**
 * parses a timecount expression into h, min, s, ms
 */
static
void parse_timecount (int *h, int *min, int *s, int *ms, char expr[])
{
	if (DEBUG_PARSETIME) report("parse_timecount starting", REP_DEBUG);
	
	int unit = PT_SECONDS;	/** the default unit is seconds unless other is specified */
	int num1 = 0, num2 = 0;
	char *n1, *n2;
	int i = strlen(expr); i--;
	
	/** parsing a letter. this should specify the time-unit */
	if (isletter(expr[i]))
	{
		/** setting unit to the correct unit and adjusts i so that we can begin parsing the number */
		switch (expr[i])
		{
			case 's':
				if (expr[i-1] == 'm') /** ms */
				{	
					i--;
					unit = PT_MILLISECONDS;
				}
				else
					unit = PT_SECONDS;	/** s */
			break;

			case 'n':				/** min */
				unit = PT_MINUTES;
				i -= 2;
			break;

			case 'h':				/** h */
				unit = PT_HOURS;
			break;
		}
	}

	/** parsing a number. this can be a whole number or two numbers separated by a "." */
	n1 = strtok (expr, ".");
	n2 = strtok (NULL, "\0");

	num1 = atoi (n1);
	if (n2 != NULL)
	{
		num2 = atoi (n2);
	}

	/** converting our num1 units and num2 fractions
	 * to mad_timer_t (seconds and fractions) */
	switch (unit)
	{
		case PT_MILLISECONDS:
			*ms = num1;
			break;
		case PT_SECONDS:
			*s = num1;		*ms = num2;
			break;
		case PT_MINUTES:
			*min = num1;	*s = num2;
			break;
		case PT_HOURS:
			*h = num1;		*min = num2;
			break;
		default:
			break;
	}
	
	if (DEBUG_PARSETIME)
	{
		char debug[STRLEN];
		sprintf (debug, "parse_timecount: %dh %dm %ds %dms", *h, *min, *s, *ms);
		report(debug, REP_DEBUG);
	}
}


/**
 * parses a partial_clock expression into h, min, s, ms
 */
static
void parse_partial_clock (int *h, int *min, int *s, int *ms, char *expr) {
	char *ptr;

	/** Parsing expr to strtok() seems to cause segfault.
	 * (is it null-terminated correctly?) Theese three lines
	 * seems to solve the problem (for now). */
//	int strl = strlen(expr);
//	char expr2[strl];
//	strncpy (expr2, expr, strl);
	
	ptr = strtok (expr, ":");	/** minutes */
	*min = atoi (ptr);
	
	ptr = strtok (NULL, ".");
	*s = atoi (ptr);
	
	if ((ptr = strtok (NULL, ".")))	/** fractions */
		*ms = atoi (ptr);
	
	if (DEBUG_PARSETIME)
	{
		char debug[STRLEN];
		sprintf (debug, "%d h, %dm, %ds, %df", *h, *min, *s, *ms);
		report(debug, REP_DEBUG);
	}
}


/**
 * parses a full_clock expression into h, min, s, ms
 */
static
void parse_full_clock (int *h, int *min, int *s, int *ms, char *expr) {
	char *ptr;

	/** Parsing expr to strtok() seems to cause segfault.
	 * (is it null-terminated correctly?) Theese three lines
	 * seems to solve the problem (for now). */
//	int strl = strlen(expr);
//	char expr2[strl];
//	strncpy (expr2, expr, strl);

	ptr = strtok (expr, ":");	/** hours */
	*h = atoi (ptr);


	ptr = strtok (NULL, ":");	/** minutes */
	*min = atoi (ptr);

	ptr = strtok (NULL, ":.");	/** seconds */
	*s = atoi (ptr);
	
	if ((ptr = strtok (NULL, ".")))	/** optional fractions */
		*ms = atoi (ptr);

	if (DEBUG_PARSETIME)
	{
		char debug[STRLEN];	
		sprintf (debug, "parse_full_clock: %dh, %dm, %ds, %dms", *h, *min, *s, *ms);
		report(debug, REP_DEBUG);
	}
}


/**
 * converts h, min, s, ms to a mad_timer_t type
 */
static
mad_timer_t convert_mad_timer_t (int *h, int *min, int *s, int *ms)
{
	mad_timer_t ret;
	ret.fraction = 0;	/** theese should already be set to 0, but let's do it explicitly too */
	ret.seconds = 0;

	if (*h)	/** if hours are a value other than 0 */
		ret.seconds += (long) (*h * 60 * 60);	/** converting hours to seconds */
	if (*min)
		ret.seconds += (long) (*min * 60);		/** converting minutes to seconds */
	if (*s)
		ret.seconds += (long) (*s);				/**	adding seconds */
	if (*ms)
		/** and converting to mad_timer_t.fractions */
		ret.fraction += (unsigned long) (*ms * MAD_TIMER_RESOLUTION / 1000);

	/** and do some correcting of the mad_timer_t */
	if (ret.fraction > MAD_TIMER_RESOLUTION)
	{
		ret.seconds  += ret.fraction / MAD_TIMER_RESOLUTION;
		ret.fraction %= MAD_TIMER_RESOLUTION;
	}
	
	if (DEBUG_PARSETIME)
	{
		char debug[STRLEN];
		sprintf (debug, "convert: %d, %d, %d, %d", *h, *min, *s, *ms);
		report(debug, REP_DEBUG);
	}
	return ret;
}


/**
 * converts h, min, s, ms to a milliseconds
 */
static
unsigned
long
int convert_milliseconds (int *h, int *min, int *s, int *ms)
{
	unsigned long int msec = 0;

	if (*h)	/** if hours are a value other than 0 */
		msec += (unsigned long) (*h * 3600000.0);	/** converting hours to milliseconds */
	if (*min)
		msec += (unsigned long) (*min * 60000.0);	/** converting minutes to milliseconds */
	if (*s)
		msec += (unsigned long) (*s * 1000.0);		/**	converting seconds to milliseconds */
	if (*ms)
		msec += (unsigned long) (*ms);
	
	if (DEBUG_PARSETIME)
	{
		char debug[STRLEN];
		sprintf (debug, "convert: %d, %d, %d, %d", *h, *min, *s, *ms);
		report(debug, REP_DEBUG);
	}
	return msec;
}


/**
 * Translated valid DAISY time formats to seconds and fractions
 * Ref: sec. 7.7, p41 in ANSI/NISO Z39.86.2005
 */
mad_timer_t parse_time_mad_time_t (char *expr)
{
	int h = 0, min = 0, s = 0, ms = 0;
	
	if (DEBUG_PARSETIME) report ("parse_time starting", REP_DEBUG);
	
	switch (type (expr))
	{
		case PT_FULL_CLOCK:
			/** Full-clock value (hours, minutes, seconds, fractions of a second) */
			parse_full_clock (&h, &min, &s, &ms, expr);
			break;
		case PT_PARTIAL_CLOCK:
			/** Partial-clock value (minutes, seconds, fractions of a second)*/
			parse_partial_clock (&h, &min, &s, &ms, expr);
			break;
		case PT_TIMECOUNT:
			/** Timecount value (one or more digits, plus an optional fraction and unit of measurement -- h=hours, min=minutes, s=seconds, ms=milliseconds) */
			parse_timecount (&h, &min, &s, &ms, expr);
			break;
		case PT_INVALID_FORMAT:
			break;
	}
	
	if (DEBUG_PARSETIME)
	{
		char debug[STRLEN];
		sprintf (debug, "parse_time: %d, %d, %d, %d", h, min, s, ms);
		report(debug, REP_DEBUG);
	}
	/** constructs a mad_timer_t struct */
	return convert_mad_timer_t (&h, &min, &s, &ms);
}


/**
 * Translated valid DAISY time formats to microseconds
 * Ref: sec. 7.7, p41 in ANSI/NISO Z39.86.2005
 */
unsigned
long
int parse_time_milliseconds (char *expr)
{
	int h = 0, min = 0, s = 0, ms = 0;
	
	if (DEBUG_PARSETIME) report ("parse_time starting", REP_DEBUG);
	
	switch (type (expr))
	{
		case PT_FULL_CLOCK:
			/** Full-clock value (hours, minutes, seconds, fractions of a second) */
			parse_full_clock (&h, &min, &s, &ms, expr);
			break;
		case PT_PARTIAL_CLOCK:
			/** Partial-clock value (minutes, seconds, fractions of a second)*/
			parse_partial_clock (&h, &min, &s, &ms, expr);
			break;
		case PT_TIMECOUNT:
			/** Timecount value (one or more digits, plus an optional fraction and unit of measurement -- h=hours, min=minutes, s=seconds, ms=milliseconds) */
			parse_timecount (&h, &min, &s, &ms, expr);
			break;
		case PT_INVALID_FORMAT:
			break;
	}
	
	if (DEBUG_PARSETIME)
	{
		char debug[STRLEN];
		sprintf (debug, "parse_time: %d, %d, %d, %d", h, min, s, ms);
		report(debug, REP_DEBUG);
	}
	/** constructs a mad_timer_t struct */
	return convert_milliseconds(&h, &min, &s, &ms);
}


/**
 * Figures out what type of time-format we're dealing with.
 */
static
enum pt_type type (char* expr) {
	char *ptr;
	int occurences = 0;
	ptr = strpbrk (expr, ":");
	while (ptr != NULL && occurences < 100)
	{
		occurences++;
		ptr = strpbrk (ptr+1, ":");
	}
	char error[STRLEN];	/** in case we need to print out an errormessage in the switch */
	switch (occurences) {
		case 2	:		/** Full-clock value */
			return PT_FULL_CLOCK;
		case 1	:		/** Partial-clock value */
			return PT_PARTIAL_CLOCK;
		case 0	:		/** Timecount-clock value */
			return PT_TIMECOUNT;
		default	:		/** Illegal format */
			snprintf (error, sizeof (error), "parsetime tried to evaluate an invalid time-format. (%s)", expr);
			report (error, REP_ERROR);
			return PT_INVALID_FORMAT;
	}
}


/**
 * Parses start- and stop-times and returns the duration (in ms)
 */
unsigned
long
int parse_time_get_dur (char *start, char *stop)
{
	unsigned long int start_ms, stop_ms;
	start_ms = parse_time_milliseconds(start);
	stop_ms = parse_time_milliseconds(stop);
	return (stop_ms - start_ms);
}
